/***************************************
 * LINITG.C / LINITG.X   Version 1.1.0
 * von Claas Lichtenberg und Zwyrd
 * zum Berechnen von Mandelbrot- und Julia-Mengen fr das ATARI-Kaosmul.Prg.
 * Das Programm ist FREIE SOFTWARE und unterliegt der GNU-GPL.
 * THIS PROGRAM COMES WITH ABSOLUTELY NO WARRANTY.
 * (C)Copyleft 11/99-1/4 Claas Lichtenberg/Zwyrd
 * Bedienungsanleitung und weiteres in der Datei LIN_ITG.TXT.
 *
 * Fehlercodes: 1 -  Benutzerabbruch
 *              2 -  Datei konnte nicht zum Lesen geoeffnet werden
 *              3 -  falsche Dateiendung
 *              4 -  falsche Dateilaenge
 *              5 -  Datei konnte nicht zum Schreiben geoeffnet werden
 *              6 -  Speicher kann nicht alloziert werden
 *              7 -  Bild ist schon fertig berechnet
 *		8 -  Falsche Fraktalmenge
 *
 * last change 5.1.4
 **************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>


#define DEBUG

#ifdef DEBUG
#define PRINTF(s1,s2)   printf(s1,s2);fflush(NULL)
#else
#define PRINTF(s1,s2)   /* */
#endif

#ifdef __unix__
#define SLASH           "/"
#else
#define SLASH           "\\"
#endif


typedef long int LINT;      /* Variable mit GENAU 32 Bit */
typedef short pixel;        /* Variable mit GENAU 16 Bit */


char    pfad[1000];
char    name[1000];
char    extension[10];
char    komplett[1000] = "";
char    ftabelle[256] = "";
double  rzahl[16];
int     datei, tiefer = 0, ftab = 0, sprung = 0;
int     fatab[258];
char    *temp;
char    buffer[512];
double  altiter;
short   format = 2, menge, param = 0;
long    *bildptr = NULL;


/*   Prototypen   */

int manuelle_eingabe (void);
void datei_eingabe (void);
void berechne (void);
int speichern (int a);
void schreibe_kdt (void);
void schreibe_itg (void);
void schreibe_bmp (void);
int min (int x, int y);
void fehler (int err);
short swap_short (short sh);
long swap_long (long lo);
void farb_tabellen (int i);


int main (int argc, char *argv[])
{
	int     rtc = 0, antw, ende = 0;
	char auswahl[10];

	PRINTF ("*********** linitg.x v110 gestartet ...\n",1);

	printf ("Programm zur Berechnung von uebergrossen Apfelmaennchen\n\n");

	if (argc > 1)
	{
		if (argc == 3)
		{
			strcpy (ftabelle, argv[2]);
			ftab = 1;
		}
		strcpy (komplett, argv[1]);
		strcpy (name, komplett);
		param = 1;
		datei_eingabe ();
	}
	else
	{
		printf ("Pfad       : "); scanf ("%s", pfad);
		printf ("Name       : "); scanf ("%s", name);
		printf ("Iterationen: "); scanf ("%lf", &rzahl[10]);

		strcpy (komplett, pfad);
		strcat (komplett, SLASH);
		strcat (komplett, name);

		if (rzahl[10] > 0)     /*-----manuelle Eingabe-------*/
		{
			antw = manuelle_eingabe ();
		}
		else /* ----------------Eingabe aus Datei------------------ */
		{
			datei_eingabe ();
		}
	}
	/* ------------------------------------------------------------ */

	if (sprung == 0)
	{
		printf ("Menge         -  %d\n", (int)rzahl[ 0]);
		printf ("Real Min      -  %.20lf\n", rzahl[ 1]);
		printf ("Real Max      -  %.20lf\n", rzahl[ 2]);
		printf ("Imag Min      -  %.20lf\n", rzahl[ 3]);
		printf ("Imag Max      -  %.20lf\n", rzahl[ 4]);
		printf ("Julia X       -  %.20lf\n", rzahl[ 5]);
		printf ("Julia Y       -  %.20lf\n", rzahl[ 6]);
		printf ("Gre X       -  %d\n", (int)rzahl[ 7]);
		printf ("Gre Y       -  %d\n", (int)rzahl[ 8]);
		printf ("Zhler        -  %d\n", (int)rzahl[ 9]);
		printf ("Iterationen   -  %d\n", (int)rzahl[10]);
		printf ("Modulo        -  %d\n", (int)rzahl[11]);
		printf ("Farben        -  %d\n", (int)rzahl[12]);
		printf ("Gre in Byte -  %d\n", (int)rzahl[13]);
		printf ("Format        -  %d\n", format);

		berechne ();

		printf ("Optionen:\n");
		printf ("i  -  ITG speichern\n");
		printf ("d  -  ITG und Bitmap speichern\n");
		printf ("b  -  Bitmap speichern\n");
		printf ("q  -  Ende\n");
		printf ("Auswahl: "); scanf ("%1s", auswahl);

		antw = toupper ((int)auswahl[0]);
	}
	else if (antw == 75)          /* .kdt-Eingabe */
		antw = 75;
	else
		antw = 66;

	if (antw == 66 || antw == 68 || antw == 73 || antw == 75) 
		speichern (antw);

	if (antw != 75)
		free (bildptr);

	PRINTF ("*********** linitg.x v110 beendet ...\n",1);

	return rtc;
}


int manuelle_eingabe (void)
{
	int     abfrage;

	/* existiert Datei schon? Wenn ja, umbenennen, loeschen, abbrechen, ... */
	do
	{
		if ((datei=open(komplett, O_RDONLY)) != -1)
		{ 									/* Datei existiert schon */
			close (datei);
			printf ("Datei existiert schon. 1=Abbruch, 2=Umbenennen, 3=Loeschen? ");
			scanf ("%d", &abfrage);

			switch (abfrage)
			{
				case 2 : printf ("Name       :"); scanf ("%s", name);
				         strcpy (komplett, pfad);
				         strcat (komplett, SLASH);
				         strcat (komplett, name);
				         break;
				case 3 : abfrage = 3;    /* existierende Datei ueberschreiben */
				         break;
				default : fehler (1);
			}
		}
		else
			abfrage=3;      /* Schleife verlassen und weiter im Text */
	} while (abfrage!=3);

	format = 2;
	if (rzahl[10] > 32767) format = 4;

	printf ("Menge (0=Mandel, 1=Julia): "); scanf ("%d", &menge);
	if (menge != 1)
	{
		menge = 0;
		rzahl[5] = 0;
		rzahl[6] = 0;
	}
	else
	{
		printf ("Julia X: "); scanf ("%lf", &rzahl[5]);
		printf ("Julia Y: "); scanf ("%lf", &rzahl[6]);
	}

	printf ("RealMin: "); scanf ("%lf", &rzahl[1]);
	printf ("RealMax: "); scanf ("%lf", &rzahl[2]);
	printf ("ImagMin: "); scanf ("%lf", &rzahl[3]);
	printf ("ImagMax: "); scanf ("%lf", &rzahl[4]);
	printf ("Groee X: "); scanf ("%lf", &rzahl[7]);
	printf ("Groee Y: "); scanf ("%lf", &rzahl[8]);

	rzahl[9] = 0; /* Zhler */
	rzahl[0] = (double)menge;

	printf ("Berechnen (=0) oder als *.KDT speichern (=1): "); scanf ("%d", &abfrage);
	if (abfrage == 1)
	{
		printf ("Modulo: "); scanf ("%lf", &rzahl[11]);
		printf ("Farben: "); scanf ("%lf", &rzahl[12]);
		sprung = 2;
		return 75;
	}
	else
	{
		rzahl[11] = 254; /* Modulo */
		rzahl[12] = 254; /* Farben */

		if (format == 4)
		{
			rzahl[13] = rzahl[7]*rzahl[8]*sizeof(LINT);
		}
		else
		{
			rzahl[13] = rzahl[7]*rzahl[8]*sizeof(pixel);
		}

		/* Speicher bereitstellen */
		if ((bildptr = (long*)malloc(rzahl[13])) == NULL)
			fehler (6);
	}
	return 0;
}

void datei_eingabe (void)
{
	char    tempstr[100];
	long	iterlim;
	int     j, a;
	double  tempdbl, dbldum;
	ssize_t anzahl;
	short	ind;

	/* Datei laden */
	if ((datei=open(komplett, O_RDONLY)) == -1)
		fehler (2);

	strcpy (extension, &name[strlen(name)-3]);
	for (temp=extension ; *temp ; temp++)
		*temp = toupper (*temp);

	if ((strcmp(extension, "ITG\0") != 0)
	&&  (strcmp(extension, "KAO\0") != 0)
	&&  (strcmp(extension, "KDT\0") != 0))
	{
		fehler (3);
	}

	anzahl = read (datei, buffer, 320);       /* Dateiheader einlesen */

	temp = (void*)buffer;

	ind = swap_short(*(pixel*)(temp)); temp +=2;

	rzahl[ 0] = swap_long(*(LINT*)(temp)); temp += sizeof (LINT);
	menge = (int)rzahl[0];
	if (menge > 1)
		fehler (8);

	rzahl[ 7] = swap_long(*(LINT*)(temp)); temp += sizeof (LINT);
	rzahl[ 8] = swap_long(*(LINT*)(temp)); temp += sizeof (LINT);
	rzahl[ 9] = swap_long(*(LINT*)(temp)); temp += sizeof (LINT);
	rzahl[10] = swap_long(*(LINT*)(temp)); temp += sizeof (LINT);
	rzahl[11] = swap_long(*(LINT*)(temp)); temp += sizeof (LINT);
	rzahl[12] = swap_long(*(LINT*)(temp)); temp += sizeof (LINT);

	if (rzahl[10] > SHRT_MAX && (ind == 0x4934 || extension[0] == 'K'))
	{
		format = 4;
		iterlim = LONG_MAX;
		rzahl[13] = rzahl[7]*rzahl[8]*sizeof(LINT);
	}
	else
	{
		format = 2;
		iterlim = SHRT_MAX;
		rzahl[13] = rzahl[7]*rzahl[8]*sizeof(pixel);
	}
	dbldum = rzahl[13];

	if (extension[0] == 'K')
	{
		temp += 20;
		rzahl[9] = 0;
	}

	tempstr[20] = '\0';
	for (j=1 ; j<7 ; j++)
	{
		memcpy (tempstr, temp, 20);
		sscanf (tempstr, "%lf", &rzahl[j]);
		temp += 20;
	}

		/* Speicher bereitstellen */
	if ((bildptr = (long*)malloc(rzahl[13])) == NULL)
		fehler (6);

	if (extension[0] == 'I')
	{
		if (rzahl[9] == rzahl[8])
		{
			if (rzahl[10] < SHRT_MAX || format == 4)
			{
				printf ("*** Bild ist fertig berechnet.\n");
				printf ("h  -  Hoehere Iterationstiefe eingeben?\n");
				printf ("b  -  Bitmap speichern\n");
				printf ("q  -  Ende\n");
				printf ("Auswahl: ");scanf ("%1s", tempstr);

				a = toupper ((int)tempstr[0]);

				if (a == 72)
				{
					tiefer = 1;
					altiter = rzahl[10];
					do
					{
						printf ("neue Iterationstiefe: ");
						scanf ("%lf", &rzahl[10]);
					} while ((rzahl[10] <= altiter) || (rzahl[10] > iterlim));
					rzahl[9] = 0;
				}
				else if (a == 66)
					sprung = 1;	/* Berechnung berspringen */
				else
					fehler (7);        /* Bild ist schon fertig berechnet */
			}
			else
				fehler (7);
		}

		lseek (datei, 150, SEEK_SET);
		tempdbl = read (datei, (void*)bildptr, dbldum);
		if (dbldum != tempdbl)
			fehler (4);
	}
	close (datei);
}


void berechne (void)
{
	int     xmax, ymax;   /* maximale Aufloesung an Pixeln */
	int     x, y;         /* Bildschirmkoordinaten */
	int     compute;
	long    inttmp;
	double  dx, dy,       /* Umrechnung von Pixelkoordinaten auf 
Mandelbrotkoordinaten */
	        mxmax, mymax, /* Mandelbrotkoordinaten */
	        mxmin, mymin, /* Mandelbrotkoordinaten */
	        mx, my,       /* Kopien fuer Mandelbrotkoordinaten */
	        mx2, my2,     /* Kopien fuer Mandelbrotkoordinaten */
	        mx3, my3,     /* Kopien fuer Mandelbrotkoordinaten */
	        xjul, yjul;   /* Juliamengenkoordinaten */
	long    imax,         /* maximale Anzahl an Iterationen */
	        zaehler,      /* Iterationszaehler */
		*lngptr;
	pixel	*pixptr;      /* Hilfzeiger fuer Pixmap */

	PRINTF ("*********** berechne() gestartet ...\n",1);

	xmax = (int)rzahl[7];
	ymax = (int)rzahl[8];

	mxmin = rzahl[1];
	mxmax = rzahl[2];

	mymin = rzahl[3];
	mymax = rzahl[4];

	xjul = rzahl[5];
	yjul = rzahl[6];


	dx = (mxmax - mxmin) / xmax;
	dy = (mymax - mymin) / ymax;

	imax = (long)rzahl[10];

	pixptr = (pixel*)((int)(bildptr) + (int)(rzahl[9] * xmax));
	lngptr = (long*)((int)(bildptr) + (int)(rzahl[9] * xmax));

	y=(int)rzahl[9];
	for (my=mymin+y*dy ; y<ymax ; y++)
	{
		printf ("\033[1G%d", y); fflush (NULL);
		x=0;
		for (mx=mxmin ; x<xmax ; x++)
		{
			compute = 1;
			if (tiefer != 0)			/*wenn  tiefer gerechnet werden soll*/
			{
				if (format == 2)
				{
					inttmp = swap_short (*pixptr);
				}
				else
				{
					inttmp = swap_long (*lngptr);
				}

				if (inttmp < altiter)
					compute = 0;
			}

			if (compute == 1)
			{
				mx2 = mx;
				my2 = my;

				mx3=0; my3=0;
				for ( zaehler=0 ; ((mx3+my3)<=4) && (zaehler<imax) ; zaehler++ )
				{
					mx3=mx2*mx2;
					my3=my2*my2;

					if ((mx3+my3)<=4)
					{
						if (menge != 0)
						{
							my2 = mx2*my2*2 + yjul;
							mx2 = mx3-my3   + xjul;
						}
						else
						{
							my2 = mx2*my2*2 + my;
							mx2 = mx3-my3   + mx;
						}
					}

				} /* Iterationsschleife */

				if (format == 2)
				{
					*pixptr = swap_short (zaehler);
				}
				else
				{
					*lngptr = swap_long (zaehler);
				}
			}
			pixptr++;
			lngptr++;

			mx+=dx;
		} /* Ende der x-Schleife */

		my+=dy;
	} /* Ende der y-Schleife */
	rzahl[9] = (double)y;
	putchar ('\n');
	PRINTF ("*********** berechne() beendet ...\n",1);
}


int speichern (int a)
{
	if (param == 0)
	{
		strcpy (komplett, pfad);
		strcat (komplett, SLASH);
		name[strlen(name)-3] = '\0';
		strcat (komplett, name);
	}
	else
	{
		komplett[strlen(komplett)-3] = '\0';
	}

	if (a == 75)
	{
		strcat (komplett, "kdt\0");
		schreibe_kdt ();
	}
	else
	{
		strcat (komplett, "itg\0");

		if (a == 68 || a == 73)
			schreibe_itg ();

		if (a == 68 || a == 66)
		{
			komplett[strlen(komplett)-3] = '\0';
			strcat (komplett, "bmp\0");

			schreibe_bmp ();
		}
	}
	return 0;
}

void schreibe_kdt (void)
{
	/* char tempstr[1000]; */
	char *ptrhelp;
	long lo;
	int  i;
	double  xzahl[10];


	PRINTF ("*********** schreibe_kdt() gestartet ...\n",1);

	i = ((rzahl[7] - 1 / 16) + 1) * 16;
	lo = ((rzahl[7] - 1 / 16) + 1) * 2 * 8 * rzahl[8];
	xzahl[0] = 8;		/* 8 bit */
	xzahl[1] = i;		/* breite in ganzen 16 */
	xzahl[2] = 0;
	xzahl[3] = 0;
	xzahl[4] = 1;		
	xzahl[5] = 32000;
	xzahl[6] = 0;
	xzahl[7] = 0;
	xzahl[8] = 0;
	xzahl[9] = lo;		/* bildgre bei 8 bit */


	ptrhelp=buffer;

	*ptrhelp = 4; ptrhelp++;         /* KaosFilm-Version 421  */
  	*ptrhelp = 33; ptrhelp++;

	lo=swap_long((long)rzahl[0]);
	memcpy ((void*)ptrhelp, (void*)&lo, sizeof (lo));
	ptrhelp+=sizeof(lo);

	for (i=7 ; i<13 ; i++)
	{
		lo=swap_long((long)rzahl[i]);
		memcpy ((void*)ptrhelp, (void*)&lo, sizeof (lo));
		ptrhelp+=sizeof(lo);
	}
	for (i=0 ; i<4 ; i++)
	{
		lo=swap_long((long)xzahl[i]);
		memcpy ((void*)ptrhelp, (void*)&lo, sizeof (lo));
		ptrhelp+=sizeof(lo);
	}
	ptrhelp+=sizeof(lo);


	for (i=1 ; i<7 ; i++)
	{
		sprintf (ptrhelp, "%-20.20lf", rzahl[i]);
		ptrhelp+=20;
	}
	for (i=4 ; i<10 ; i++)
	{
		sprintf (ptrhelp, "%-20.20lf", xzahl[i]);
		ptrhelp+=20;
	}

	ptrhelp+=22;

	sprintf (ptrhelp, "%-8s", name);
	ptrhelp+=8;


	if ((datei=open (komplett, O_WRONLY | O_CREAT | O_TRUNC, 
S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH))==-1)
		fehler (5);

	write (datei, buffer, (int)ptrhelp-(int)buffer);

	close (datei);

	PRINTF ("*********** schreibe_kdt() beendet ...\n",1);

}

void schreibe_itg (void)
{
	/* char tempstr[1000]; */
	char *ptrhelp;
	long lo;
	int  i;

	PRINTF ("*********** schreibe_itg() gestartet ...\n",1);

	ptrhelp=buffer;

	*ptrhelp = 73; ptrhelp++;         /* I2/I4-kennung  */
  	*ptrhelp = 48 + format; ptrhelp++;

	lo=swap_long((long)rzahl[0]);
	memcpy ((void*)ptrhelp, (void*)&lo, sizeof (lo));
	ptrhelp+=sizeof(lo);

	for (i=7 ; i<13 ; i++)
	{
		lo=swap_long((long)rzahl[i]);
		memcpy ((void*)ptrhelp, (void*)&lo, sizeof (lo));
		ptrhelp+=sizeof(lo);
	}

	for (i=1 ; i<7 ; i++)
	{
		sprintf (ptrhelp, "%-20.20lf", rzahl[i]);
		ptrhelp+=20;
	}

	if ((datei=open (komplett, O_WRONLY | O_CREAT | O_TRUNC, 
S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH))==-1)
		fehler (5);

	write (datei, buffer, (int)ptrhelp-(int)buffer);
	write (datei, bildptr, rzahl[13]);

	close (datei);

	PRINTF ("*********** schreibe_itg() beendet ...\n",1);
}


void schreibe_bmp (void)
{
	char		kcopfad[256] = "/usr/share/linitg/";
	char		buff[10], xx[5];
	char		bm[3] = "BM";
	unsigned char	by0 = 0, co0, co1, co2, pz;
	int		fx, fx2, fx3, fy, fy2;
	int		zahl[5] = {0,0,0,0,0};
	int		z = 0, u = 0, c, sw = 0;
	int		abfrage;
	short		sh = 0, *hdoptr;
	short		umre[16] = {0,2,3,6,4,7,5,8,9,10,11,14,12,15,13,1};
	short		umpi[16] = {0,15,1,2,4,6,3,5,7,8,9,10,12,14,11,13};
	long		lo = 0, bmv, bmv15, v = 40, i, os;
	long		inttmp, imax, modu;
	long		*bmptr, *hdptr, *lngptr;
	FILE		*kco;
	pixel		*pixptr;
	double	mcol;

	PRINTF ("*********** erstelle bitmap ...\n",1);


	rzahl[14] = (((int)rzahl[7] - 1) / 4 + 1) * 4;
	rzahl[15] = rzahl[14] * rzahl[8];

			/*speicher fuer Bitmap bereitstellen*/
	if ((bmptr = (long*)malloc(rzahl[15])) == NULL)
		fehler (6);

	if (ftab == 0)
	{
		printf ("Dateieingabe -  0\n");
		printf ("Kreis        -  1\n");
		printf ("Pastel 1     -  2\n");
		printf ("Pastel 2     -  3\n");
		printf ("Rasta 1      -  4\n");
		printf ("Rasta 2      -  5\n");
		printf ("Rasta 3      -  6\n");
		printf ("Bunt         -  7\n");
		printf ("Grau/Gruen   -  8\n");
		printf ("Rot          -  9\n");
		printf ("Gelb         - 10\n");
		printf ("Gruen        - 11\n");
		printf ("Blau         - 12\n");
		printf ("Lila         - 13\n");
		printf ("Grau         - 14\n");
		printf ("Schwarz/Gelb - 15\n");
		printf ("Gelb/Schwarz - 16\n");


		printf ("Farbtabelle: "); scanf ("%d", &abfrage);
		if (abfrage == 0)
		{
			printf ("Datei: "); scanf ("%255s", ftabelle);
			ftab = 1;
		}
		else
			farb_tabellen (abfrage);
	}
	if (ftab != 0)
	{
		if (strcmp(ftabelle, ".kco\0") != 0)
			strcat (ftabelle, ".kco");

		if ((kco = fopen (ftabelle, "r")) == NULL)
		{
			strcat (kcopfad, ftabelle);
			if ((kco = fopen (kcopfad, "r")) == NULL)
				fehler (2);
		}

		while ((c = fgetc (kco)) != EOF)		/*farbtabelle auslesen*/
		{
			if (c > 43)
			{
				if (c == ',')
				{
					while (u < 4)
					{
						for (i=u-1; i>=0; i--)
							zahl[i+1] = zahl[i];
						zahl[0] = 0;
						u++;
					}
					fatab[z] = zahl[0]*1000 + zahl[1]*100 + zahl[2]*10 + zahl[3];
					z++;
					u = 0;
					for (i=0; i<4; i++) zahl[i] = 0;
				}
				else
				{
					zahl[u] = c - 48;
					u++;
				}
			}
		}

		fclose (kco);
	}

	printf ("Modulo = %d. Ist das ok? y/n: ",(int)rzahl[11]);
	scanf ("%s", buff);
	if (toupper ((int)buff[0]) == 'N')
	{
		printf ("Modulo: ");
		scanf ("%lf", &rzahl[11]);
	}

	printf ("Farben = %d. Ist das ok? y/n: ",(int)rzahl[12]);
	scanf ("%s", buff);
	if (toupper ((int)buff[0]) == 'N')
	{
		printf ("Farben: ");
		scanf ("%lf", &rzahl[12]);
	}

	if (rzahl[12] > 254)
		rzahl[12] = 254;
	if (rzahl[12] < 2)
		rzahl[12] = 2;
	if (rzahl[11] < 2)
		rzahl[11] = 2;

	printf ("Hintergrund Schwarz. Ist das ok? y/n: ");
	scanf ("%s", buff);
	if (toupper ((int)buff[0]) == 'N')
	{
		printf ("S / W: ");
		scanf ("%1s", buff);
	}

	if (toupper ((int)buff[0]) == 'W')
		sw = 1;

	bmv = 1078;				/*bitmap-header erstellen*/
	bmv15 = bmv + (long)rzahl[15];

	if ((hdptr = (long*)malloc(bmv)) == NULL)
		fehler (6);

	hdoptr = (pixel*)hdptr;

	memcpy (hdoptr, bm, 2); hdoptr++;
	memcpy (hdoptr, (void*)&bmv15, 4); hdoptr++; hdoptr++;
	memcpy (hdoptr, &sh, 2); hdoptr++;
	memcpy (hdoptr, &sh, 2); hdoptr++;
	memcpy (hdoptr, &bmv, 4); hdoptr++; hdoptr++;
	memcpy (hdoptr, &v, 4); hdoptr++; hdoptr++;
	v = (long)rzahl[7];
	memcpy (hdoptr, (void*)&v, 4); hdoptr++; hdoptr++;
	v = (long)rzahl[8];
	memcpy (hdoptr, (void*)&v, 4); hdoptr++; hdoptr++;
	sh = 1;
	memcpy (hdoptr, &sh, 2); hdoptr++;
	sh = 8;
	memcpy (hdoptr, &sh, 2); hdoptr++;
	memcpy (hdoptr, &lo, 4); hdoptr++; hdoptr++;
	v = (long)rzahl[15];
	memcpy (hdoptr, &v, 4); hdoptr++; hdoptr++;
	memcpy (hdoptr, &lo, 4); hdoptr++; hdoptr++;
	memcpy (hdoptr, &lo, 4); hdoptr++; hdoptr++;
	v = 256;
	memcpy (hdoptr, &v, 4); hdoptr++; hdoptr++;
	memcpy (hdoptr, &v, 4); hdoptr++; hdoptr++;

	c = 0;
	for (i=0; i<256; i++)
	{
		c = i;
		if (i < 16)
			c = umre[i];

		co2 = fatab[c] / 256;
		co1 = (fatab[c] - co2 * 256) / 16;
		co0 = fatab[c] - co2 * 256 - co1 * 16;
		co2 = co2 *16 + min (co2, 1) * 15;
		co1 = co1 *16 + min (co1, 1) * 15;
		co0 = co0 *16 + min (co0, 1) * 15;
		memset ((void*)xx, co2, 1);
		memset ((void*)(xx + 1), co1, 1);
		memset ((void*)(xx + 2), co0, 1);
		memset ((void*)(xx + 3), by0, 1);
		memmove (hdoptr, xx, 4);
		hdoptr++;
		hdoptr++;
	}

	imax = (long)rzahl[10];			/*bitmap erstellen*/
	modu = (long)rzahl[11];
	mcol = (double)modu / rzahl[12];
	fy2 = (int)rzahl[8] - 1;
	fx2 = (int)rzahl[14];
	fx3 = (int)rzahl[7];
	pixptr = (pixel*)((int)(bildptr));
	lngptr = (long*)((int)(bildptr));

	for (fy = fy2; fy >= 0; fy--)
	{
 		os = (long)bmptr + fy * fx2;
		for (fx=0; fx < fx2; fx++)
		{
			pz = 1; 
			if (sw == 1)
				pz = 0;

			if ( fx < fx3)
			{
				if (format == 2)
				{
					inttmp = swap_short (*pixptr);
				}
				else
				{
					inttmp = swap_long (*lngptr);
				}
				pixptr++;
				lngptr++;
				if (inttmp < imax)
				{
					pz = (char)((inttmp % modu) / mcol + 2); 
				}
				if (pz < 16)
					pz = (char)umpi[pz];
			}

			memset ((void*)os, pz, 1);
			os++;
		}
	}

	if ((datei=open (komplett, O_WRONLY | O_CREAT | O_TRUNC, 
S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH))==-1)
		fehler (5);

	write (datei, hdptr, bmv);
	write (datei, bmptr, (int)rzahl[15]);

	close (datei);

	free (hdptr);
	free (bmptr);

	PRINTF ("*********** bitmap fertig ...\n",1);
}


int min (int x, int y)
{
	if (x < y) return x;
	return y;
}


void fehler (int err)
{
	switch (err)
	{
		case 1  : printf ("*** Benutzerabbruch\n"); break;
		case 2  : printf ("*** Datei konnte nicht zum Lesen geoeffnet werden.\n"); 
break;
		case 3  : printf ("*** Falsche Dateiendung\n"); break;
		case 4  : printf ("*** Falsche Dateilaenge\n"); break;
		case 5  : printf ("*** Datei konnte nicht zum Schreiben geoeffnet werden.\n"); 

break;
		case 6  : printf ("*** Speicher kann nicht alloziert werden.\n"); break;
		case 7  : printf ("*** Bild ist schon fertig berechnet.\n"); break;
		case 8  : printf ("*** Falsche Fraktalmenge\n"); break;
		default : printf ("Unbekannter Fehler #%d\n", err);
	}
	exit (err);
}


short swap_short (short sh)
{
	unsigned short  *ush;

	ush = &sh;

	return (short)(*ush<<8)+(*ush>>8);
}


long swap_long (long lo)
{
	unsigned long  *ulo;

	ulo = &lo;

	return (long)((swap_short(*ulo&0xFFFF)<<16)+swap_short(*ulo>>16));
}

void farb_tabellen (int i)
{
	int	j, z;
	int	cotabs254[2032] = {
		2056,2055,2054,2053,2052,2051,2050,2049,2048,2304,2560,2816,3072,3328,3584,3840,3856,3872,3888,3904,3920,3936,3952,3968,3984,4000,4016,4032,4048,4064,4080,3824,3568,3312,3056,2800,2544,2288,2032,1776,1520,1264,1008,752,496,240,224,208,192,176,160,144,128,144,160,176,192,208,224,240,241,242,
		243,244,245,246,247,248,249,250,251,252,253,254,255,239,223,207,191,175,159,143,127,111,95,79,63,47,31,15,14,13,12,11,10,9,8,7,8,9,10,11,12,13,14,15,271,527,783,1039,1295,1551,1807,2063,2319,2575,2831,3087,3343,3599,3855,3598,3341,3084,2827,2570,
		2313,2056,2055,2054,2053,2052,2051,2050,2049,2048,2304,2560,2816,3072,3328,3584,3840,3856,3872,3888,3904,3920,3936,3952,3968,3984,4000,4016,4032,4048,4064,4080,3824,3568,3312,3056,2800,2544,2288,2032,1776,1520,1264,1008,752,496,240,224,208,192,176,160,144,128,144,160,176,192,208,224,240,241,242,243,
		244,245,246,247,248,249,250,251,252,253,254,255,239,223,207,191,175,159,143,127,111,95,79,63,47,31,15,14,13,12,11,10,9,8,7,8,9,10,11,12,13,14,15,271,527,783,1039,1295,1551,1807,2063,2319,2575,2831,3087,3343,3599,3855,3598,3341,3084,2827,2570,2313,

		2104,2103,2102,2101,2100,2099,2098,2097,2096,2352,2609,2866,3123,3363,3606,3843,3859,3875,3891,3907,3923,3939,3955,3971,3987,4003,4019,4035,4051,4067,4083,3827,3571,3315,3059,2803,2547,2291,2035,1779,1523,1267,1011,755,499,243,483,723,963,947,931,915,899,915,931,947,963,978,993,1008,1009,1010,
		1011,1012,1013,1014,1015,1016,1017,1018,1019,1020,1021,1022,1023,1007,991,975,959,943,927,911,895,879,863,847,831,815,799,783,798,813,828,827,826,825,824,823,824,825,826,827,828,573,318,63,319,575,831,1087,1343,1599,1855,2111,2367,2623,2879,3135,3391,3647,3903,3646,3389,3132,2875,2618,
		2361,2104,2103,2102,2101,2100,2099,2098,2097,2096,2352,2609,2866,3123,3363,3606,3843,3859,3875,3891,3907,3923,3939,3955,3971,3987,4003,4019,4035,4051,4067,4083,3827,3571,3315,3059,2803,2547,2291,2035,1779,1523,1267,1011,755,499,243,483,723,963,947,931,915,899,915,931,947,963,978,993,1008,1009,1010,1011,
		1012,1013,1014,1015,1016,1017,1018,1019,1020,1021,1022,1023,1007,991,975,959,943,927,911,895,879,863,847,831,815,799,783,798,813,828,827,826,825,824,823,824,825,826,827,828,573,318,63,319,575,831,1087,1343,1599,1855,2111,2367,2623,2879,3135,3391,3647,3903,3646,3389,3132,2875,2618,2361,

		2152,2151,2150,2149,2148,2147,2146,2145,2144,2401,2642,2883,3124,3365,3606,3846,3862,3878,3894,3910,3926,3942,3958,3974,3990,4006,4022,4038,4054,4070,4086,3830,3574,3318,3062,2806,2550,2294,2038,1782,1526,1270,1014,758,502,246,486,726,966,1206,1446,1686,1670,1686,1701,1716,1731,1746,1761,1776,1777,1778,
		1779,1780,1781,1782,1783,1784,1785,1786,1787,1788,1789,1790,1791,1775,1759,1743,1727,1711,1695,1679,1663,1647,1631,1615,1599,1583,1567,1551,1550,1549,1548,1563,1578,1593,1608,1623,1640,1641,1386,1131,876,621,366,111,367,623,879,1135,1391,1647,1903,2159,2415,2671,2927,3183,3439,3695,3951,3694,3437,3180,2923,2666,
		2409,2152,2151,2150,2149,2148,2147,2146,2145,2144,2401,2642,2883,3124,3365,3606,3846,3862,3878,3894,3910,3926,3942,3958,3974,3990,4006,4022,4038,4054,4070,4086,3830,3574,3318,3062,2806,2550,2294,2038,1782,1526,1270,1014,758,502,246,486,726,966,1206,1446,1686,1670,1686,1701,1716,1731,1746,1761,1776,1777,1778,1779,
		1780,1781,1782,1783,1784,1785,1786,1787,1788,1789,1790,1791,1775,1759,1743,1727,1711,1695,1679,1663,1647,1631,1615,1599,1583,1567,1551,1550,1549,1548,1563,1578,1593,1608,1623,1640,1641,1386,1131,876,621,366,111,367,623,879,1135,1391,1647,1903,2159,2415,2671,2927,3183,3439,3695,3951,3694,3437,3180,2923,2666,2409,
		
		3,4,276,5,277,6,278,7,279,8,280,9,281,10,282,11,283,12,284,13,285,14,286,15,783,31,799,47,1071,63,1087,79,1103,95,1375,111,1391,127,1407,143,1423,159,1439,175,1455,191,1471,207,1487,223,1503,239,1519,255,511,767,1023,1279,1535,1791,2047,2303,
		2559,2815,3071,3327,3326,3582,3581,3580,3324,3323,3067,3066,2810,2809,2553,2552,2296,2295,2039,2039,1782,1781,1525,1524,1268,1267,1011,1010,754,753,497,496,240,1509,995,224,1493,979,208,1477,963,192,1461,947,176,1445,931,160,1429,915,144,1413,899,128,883,112,867,96,851,80,835,64,562,48,
		32,3,4,276,5,277,6,278,7,279,8,280,9,281,10,282,11,283,12,284,13,285,14,286,15,783,31,799,47,1071,63,1087,79,1103,95,1375,111,1391,127,1407,143,1423,159,1439,175,1455,191,1471,207,1487,223,1503,239,1519,255,511,767,1023,1279,1535,1791,2047,2303,2559,
		2815,3071,3327,3326,3582,3581,3580,3324,3323,3067,3066,2810,2809,2553,2552,2296,2295,2039,2039,1782,1781,1525,1524,1268,1267,1011,1010,754,753,497,496,240,1509,995,224,1493,979,208,1477,963,192,1461,947,176,1445,931,160,1429,915,144,1413,899,128,883,112,867,96,851,80,835,64,562,48,32,

		3,4,276,5,277,6,278,7,279,8,280,9,281,10,282,11,283,12,284,13,285,14,286,15,783,31,799,47,1071,63,1087,79,1103,95,1375,111,1391,127,1407,143,1423,159,1439,175,1455,191,1471,207,1487,223,1503,239,1519,255,511,767,1023,1279,1535,1791,2047,2303,
		2559,2815,3071,3327,3326,3582,3581,3580,3324,3323,3067,3066,2810,2809,2553,2552,2296,2295,2039,2039,1782,1781,1525,1524,1268,1267,1011,1010,754,753,497,496,240,1509,995,224,1493,979,208,1477,963,192,1461,947,176,1445,931,160,1429,915,144,1413,899,128,883,112,867,96,851,80,835,64,562,48,
		32,16,32,48,562,64,835,80,851,96,867,112,883,128,899,1413,144,915,1429,160,931,1445,176,947,1461,192,963,1477,208,979,1493,224,995,1509,240,496,497,753,754,1010,1011,1267,1268,1524,1525,1781,1782,2038,2039,2295,2296,2552,2553,2809,2810,3066,3067,3323,3324,3580,3581,3582,3326,3327,
		3071,2815,2559,2303,2047,1791,1535,1279,1023,767,511,255,1519,239,1503,223,1487,207,1471,191,1455,175,1439,159,1423,143,1407,127,1391,111,1375,95,1103,79,1087,63,1071,47,799,31,783,15,286,14,285,13,284,12,283,11,282,10,281,9,280,8,279,7,278,6,277,5,276,4,

		16,32,48,562,64,835,80,851,96,867,112,883,128,899,1413,144,915,1429,160,931,1445,176,947,1461,192,963,1477,208,979,1493,224,995,1509,240,496,497,753,754,1010,1011,1267,1268,1524,1525,1781,1782,2038,2039,2295,2296,2552,2553,2809,2810,3066,3067,3323,3324,3580,3581,3582,3326,
		3327,3071,2815,2559,2303,2047,1791,1535,1279,1023,767,511,255,1519,239,1503,223,1487,207,1471,191,1455,175,1439,159,1423,143,1407,127,1391,111,1375,95,1103,79,1087,63,1071,47,799,31,783,15,286,14,285,13,284,12,283,11,282,10,281,9,280,8,279,7,278,6,277,5,276,
		4,16,32,48,562,64,835,80,851,96,867,112,883,128,899,1413,144,915,1429,160,931,1445,176,947,1461,192,963,1477,208,979,1493,224,995,1509,240,496,497,753,754,1010,1011,1267,1268,1524,1525,1781,1782,2038,2039,2295,2296,2552,2553,2809,2810,3066,3067,3323,3324,3580,3581,3582,3326,3327,
		3071,2815,2559,2303,2047,1791,1535,1279,1023,767,511,255,1519,239,1503,223,1487,207,1471,191,1455,175,1439,159,1423,143,1407,127,1391,111,1375,95,1103,79,1087,63,1071,47,799,31,783,15,286,14,285,13,284,12,283,11,282,10,281,9,280,8,279,7,278,6,277,5,276,4,

		1024,1280,1536,1792,2048,2304,2560,2816,3072,3328,3584,3840,3856,3872,3888,3904,3920,3936,3952,3968,3984,4000,4016,4032,4048,4064,4080,4081,4082,4083,4084,4085,4086,4087,4088,4089,4090,4091,4092,4093,4094,3839,3583,3327,3071,2815,2559,2303,2047,1791,1535,1279,1023,767,511,255,239,223,207,191,175,159,
		143,127,111,95,79,63,47,31,15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,273,546,819,1092,1365,1638,1911,2184,2457,2730,3003,3276,3549,3822,3581,3324,3067,2810,2553,2296,2039,1782,1525,1268,1011,754,497,240,496,752,1008,1264,1520,1776,2032,2288,2544,2800,3056,3312,3568,
		3824,4080,4064,4048,4032,4016,4000,3984,3968,3952,3936,3920,3904,3888,3872,3856,3840,3841,3842,3843,3844,3845,3846,3847,3848,3849,3850,3851,3852,3853,3854,3855,3599,3343,3087,2831,2575,2319,2063,1807,1551,1295,1039,783,527,271,15,287,559,831,1103,1375,1647,1919,2191,2463,2735,3007,3279,3551,3823,4078,4061,4044,
		4027,4010,3993,3976,3959,3942,3925,3908,3891,3874,3857,3840,3600,3360,3120,2880,2640,2400,2160,1920,1680,1440,1200,960,720,480,240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255,495,735,975,1215,1455,1695,1935,2175,2415,2655,2895,3135,3375,3615,3855,3598,3341,3084,2827,2570,2313,2056,

		273,32,290,48,546,64,563,80,819,80,836,112,1092,128,1109,144,1365,160,1382,176,1638,192,1655,208,1911,224,1928,240,2184,241,2201,243,2457,245,2474,247,2730,249,771,2104,1028,2377,1301,2650,1574,2907,1831,3164,2104,3421,607,1439,879,1455,1151,1471,1167,1487,1439,1503,1455,1519,
		3822,546,273,32,290,48,546,64,563,80,819,80,836,112,1092,128,1109,144,1365,160,1382,176,1638,192,1655,208,1911,224,1928,240,2184,241,2201,243,2457,245,2474,247,2730,249,771,2104,1028,2377,1301,2650,1574,2907,1831,3164,2104,3421,607,1439,879,1455,1151,1471,1167,1487,1439,1503,1455,1519,
		3822,546,273,32,290,48,546,64,563,80,819,80,836,112,1092,128,1109,144,1365,160,1382,176,1638,192,1655,208,1911,224,1928,240,2184,241,2201,243,2457,245,2474,247,2730,249,771,2104,1028,2377,1301,2650,1574,2907,1831,3164,2104,3421,607,1439,879,1455,1151,1471,1167,1487,1439,1503,1455,1519,
		3822,546,273,32,290,48,546,64,563,80,819,80,836,112,1092,128,1109,144,1365,160,1382,176,1638,192,1655,208,1911,224,1928,240,2184,241,2201,243,2457,245,2474,247,2730,249,771,2104,1028,2377,1301,2650,1574,2907,1831,3164,2104,3421,607,1439,879,1455,1151,1471,1167,1487,1439,1503,1455,1519
		};
	int	cotabs32[256] = {
		4,5,6,7,8,9,10,11,12,13,14,15,31,47,63,79,95,111,127,143,159,415,671,927,1183,1439,1455,1471,1487,1503,1519,1535,
		102,118,119,135,136,152,153,169,170,186,187,203,204,220,221,237,238,254,255,511,767,1023,1279,1535,1791,2047,2303,2559,2815,3071,3327,3583,
		64,80,96,112,128,144,160,176,192,208,224,240,496,752,1008,1264,1520,1776,2032,2288,2544,2545,2546,2547,2548,2549,2805,3061,3317,3573,3829,4085,
		1024,1280,1536,1792,2048,2304,2560,2816,3072,3328,3584,3840,3856,3872,3888,3904,3920,3936,3952,3968,3984,3985,3986,3987,3988,3989,4005,4021,4037,4053,4069,4085,
		1028,1284,1285,1541,1542,1798,1799,2055,2056,2312,2313,2569,2570,2826,2827,3083,3084,3340,3341,3597,3598,3854,3855,3871,3887,3903,3919,3935,3951,3967,3983,3999,
		16,17,273,290,546,563,819,836,1092,1109,1365,1382,1638,1655,1911,1928,2184,2185,2201,2457,2458,2474,2730,2747,3003,3020,3276,3293,3549,3566,3822,3823,
		273,546,819,1092,1365,1638,1911,2184,2457,2730,3003,3276,3549,3293,3037,2781,2525,2526,2270,2286,2030,2031,1775,1791,1279,1023,767,511,1023,239,223,207,
		207,223,239,255,511,767,1023,1279,1535,1518,1774,2030,2285,2269,2525,2781,3037,3293,3565,3549,3276,3003,2730,2457,2184,1911,1638,1365,1092,819,546,273
		};

	fatab[0] = 4095;
	fatab[1] = 0;
	i-=1;

	if (i < 8)
	{
		for (j=0; j<254; j++)
		{
			fatab[j + 2] = cotabs254[i * 254 + j];
		}
	}
	else
	{
		for (z=0; z<8; z++)
		{
			for (j=0; j<32; j++)
			{
				fatab[z * 32 + j + 2] = cotabs32[(i - 8) * 32 + j];
			}
		}
	}
}


